from .config import DB_MODULES
from .base import BaseDB, IDB
import importlib
from typing import Tuple


class UnionDB:
    def __init__(self, dbs: Tuple[BaseDB]):
        self.dbs = dbs

    def exists(self, key: str):
        return any(db.exists(key) for db in self.dbs)

    def __getitem__(self, key: str):
        for db in self.dbs:
            try:
                return db[key]
            except KeyError:
                pass

        raise KeyError(key)

    def __setitem__(self, key, value):
        print(f'setting {key} to {value}')
        print(f'db = {self.dbs[0]}')
        self.dbs[0][key] = value


_db_cache = {}


def connect(url: str) -> IDB:
    dbs = [_connect(x) for x in url.split(';')]
    if len(dbs) == 1:
        return dbs[0]
    return UnionDB(dbs)


def _connect(url: str) -> BaseDB:
    global _db_cache
    if url not in _db_cache:
        db_cls = _resolve_db_class(url)
        _db_cache[url] = db_cls(url)

    return _db_cache[url]


def _resolve_db_class(url: str):
    db_type = url.split(':', 1)[0]
    assert db_type in DB_MODULES, \
        '{} is not one of the valid db types: {}'.format(
            db_type,
            list(iter(DB_MODULES.keys())))

    class_name = DB_MODULES[db_type]
    module_path = __name__.rsplit('.', 1)[0] + '.impl.' + db_type
    m = importlib.import_module(module_path)
    return getattr(m, class_name)



from .exceptions import DbObjException
from .dbobj import IS_DB_OBJ
import importlib
import pickle


DBOBJ_CONFIG_PATH = '.configs/objdb'


class ObjDBMixin:
    """
    A Mixin class that BaseDB inherit from.

    This provides functionality for decorated python objects
    to be stored in KYDB
    """

    def upload_objdb_config(self, config):
        """Upload ObjDB config to KYDB

           This should only need to be done when new classes are registered or existing ones changes path.

::

    db = kydb.connect('memory://decorated_py_obj')

    db.upload_objdb_config({
        'Greeter': {
            'module_path': 'path.to.module',
            'class_name': 'Greeter'
        }
    })
        """
        self[DBOBJ_CONFIG_PATH] = config

    def _get_dbobj_config(self, class_name: str):
        try:
            config = self[DBOBJ_CONFIG_PATH]
        except KeyError:
            raise DbObjException('Missing config file '
                                 + DBOBJ_CONFIG_PATH
                                 + 'Please ensure it exists in ' + str(self))

        cfg = config[class_name]
        if not ('module_path' in cfg and 'class_name' in cfg):
            raise DbObjException('config must have module_path '
                                 'and class_name, instead got '
                                 + str(config))

        return cfg

    def db_obj_new(self, class_name: str, key: str, kwargs: dict):
        config = self._get_dbobj_config(class_name)
        m = importlib.import_module(config['module_path'])
        cls = getattr(m, config['class_name'])
        return cls(self, class_name, key, **kwargs)

    @staticmethod
    def is_data_dbobj(data) -> bool:
        return isinstance(data, dict) and data.get(IS_DB_OBJ, False)

    @staticmethod
    def is_dbobj(obj) -> bool:
        return getattr(obj, IS_DB_OBJ, False)

    def read_dbobj(self, data):
        meta = data['meta']
        return self.db_obj_new(meta['class_name'],
                               meta['key'], data['data'])

    @staticmethod
    def write_dbobj(obj):
        data = {
            IS_DB_OBJ: True,
            'meta': {
                'key': obj.key,
                'class_name': obj.class_name,
            },
            'data': obj.get_stored_dict()
        }
        obj.db.set_raw(obj.key, pickle.dumps(data))



IS_DB_OBJ = '__is_dbobj__'


class StoredValue:
    """Any function decorated with @kydb.stored would turn the function into a StoredValue

    For example, the method ``name`` would turn into a ``StoredValue`` because of the decorator.

::    

    class Greeter(kydb.DbObj):

        @kydb.stored
        def name(self):
            return 'John'
    """

    def __init__(self, default_func):
        self._default_func = default_func
        self._value = None

    def get_default(self, obj):
        return self._default_func(obj)

    def setvalue(self, value):
        """Set the value
        :params value: the value to set

example::

    greeter = db.new('Greeter', key)
    greeter.name.setvalue('Jane')
    greeter.name() # returns 'Jane'

        """
        self._value = value

    def clear(self):
        """Clears the value.

        This sets the value back to default.

example::

    greeter = db.new('Greeter', key) # Default is 'John'
    greeter.name.setvalue('Jane')
    greeter.name() # returns 'Jane'
    greater.clear()
    greeter.name() # returns 'John'

        """
        self._value = self.get_default(self)

    def __call__(self):
        return self._value


class DbObj:
    """Derive this class to enable it for Python Object DB API

    example

::

    class Greeter(kydb.DbObj):

        def init(self):
            self.greet_count = 0

        @kydb.stored
        def name(self):
            return 'John'

        def greet(self):
            self.greet_count += 1
            return 'Hello ' + self.name()

    """

    def __init__(self, db, class_name: str, key: str, **kwargs):
        self.class_name = class_name
        self.db = db
        self.key = key

        self._stored_attrs = [x for x in dir(self) if isinstance(
            getattr(self, x), StoredValue)]

        setattr(self, IS_DB_OBJ, True)

        for attr in self._stored_attrs:
            sv = getattr(self, attr)
            v = kwargs.get(attr)
            if not v:
                v = sv.get_default(self)

            sv.setvalue(v)

        self.init()

    def init(self):
        """ Implement this to add additional initialisation

        example

::

    class MyClass(kydb.DbObj):
        def init(self):
            self._some_network_resource = get_network_resource()
        """
        pass

    def get_stored_dict(self):
        """The dictionary to be serialised and stored in DB.

        Example:

::

    class Greeter(kydb.DbObj):

        def init(self):
            self.greet_count = 0 ##  Not persisted

        @kydb.stored
        def name(self):
            return 'John'

        def greet(self):
            self.greet_count += 1
            return 'Hello ' + self.name()

    greeter.get_stored_dict() # returns {'name': 'Mary'}
    """
        return {x: getattr(self, x)() for x in self._stored_attrs}

    def write(self):
        self.db[self.key] = self

    def delete(self):
        print('++== deleteing ' + self.key)
        self.db.delete(self.key)


def stored(f):
    return StoredValue(f)



from abc import ABC
import pickle
from typing import Tuple
from .objdb import ObjDBMixin


class IDB(ABC):
    """The interface that all KYDB adheres to
    """

    def __getitem__(self, key: str):
        """Get data from the DB based on key

        To be implemented by derived class

        :param key: str:  The key to get.

example::

    db[key] # returns the object with key

        """
        raise NotImplementedError()

    def __setitem__(self, key: str, value):
        """Set data from the DB based on key
        To be implemented by derived class

        :param key: str:  The key to set.

example::

    db[key] = value # sets key to value
        """
        raise NotImplementedError()

    def delete(self, key: str):
        """
        Delete a key from the db.
        To be implemented by derived class

        :param key: str:  The key to delete.

example::

    db.delete(key) # Deletes data with key
        """
        raise NotImplementedError()

    def new(self, class_name: str, key: str, **kwargs):
        """
        Create a new object on the DB.
        The object is not persisted until obj.write() is called.

        :param class_name: str: name of the class.
                                This name must be in the config registry
        :param key: str: The key to persist on the DB
        :param kwargs: the stored attributes to set on the obj
        :returns: an obj of type defined by class_name

example::

    obj = db.new('MyClass', key, foo=3)
        """


class BaseDB(IDB, ObjDBMixin):
    """
    All implementations derives from this base class except for UnionDB.

    This class and all derived classes are never instantiated directly.

Instead use the ``connect``. i.e.::

    import kydb

    db = kydb.connect('dynamodb://my-table')
    """

    def __init__(self, url: str):
        self.db_type = url.split(':', 1)[0]
        self.db_name, self.base_path = self._get_name_and_basepath(url)
        self.url = url
        self._cache = {}

    @staticmethod
    def _get_name_and_basepath(url: str) -> Tuple[str, str]:
        """
        get the db_name and base_path

        :param url: str:
        :returns: a tuple (db_name, base_path)

        """
        base_path = '/'
        parts = url.split('/', 3)
        num_parts = len(parts)
        assert num_parts in (3, 4)

        db_name = parts[2]
        if len(parts) == 4:
            base_path += parts[-1]

        if base_path[-1] != '/':
            base_path += '/'

        return db_name, base_path

    def _get_full_path(self, key: str) -> str:
        """
        get the fullpath from key including base_path.

        :param key: str:
        :returns: The fullpath from key including base_path.

        """
        if key.startswith('/'):
            key = key[1:]

        return self.base_path + key

    def _serialise(self, obj):
        """
        Serialises the object.

        :param obj: object to serialise
        :returns: The obj pickled

        """
        return pickle.dumps(obj)

    def _deserialise(self, data):
        """
        Deserialise the data

        :param obj:
        :returns: Unpickle data
        """
        return pickle.loads(data)

    def exists(self, key) -> bool:
        """
        Check if a key exists in the DB

        :param key: the key
        :returns: True if key exists, False otherwise.

        Note that base implementation is to try and get the object
        derived db would either return the object or must raise
        KeyError if not found.

        Returns False if KeyError is raised, True otherwise

There are also side effects. The result is cached so that::

        if db.exists(key):
            db[key]

would only hit the DB once.
        """
        # TODO: defer this implmementaion to derived class as exist()
        # should not need to load the object
        try:
            self[key]
            return True
        except KeyError:
            return False

    def __getitem__(self, key: str):
        """
        Get data from the DB based on key

        :param key: str:  The key to get.

        :param key: str:  The key to get.

example::

    db[key] # returns the object with key
        """
        return self.read(key)

    def refresh(self, key=None):
        """
        Flush the cache

        :param key: Optionally choose which key to flush (Default value = None)

example::

    obj = db.new('MyClass', key)
    obj.write()
    db[key] # read from cache
    db.refresh() # Or db.refresh(key)
    db[key] # read from DB

        """
        if key:
            del self._cache[key]
        else:
            self._cache = {}

    def read(self, key: str, reload=False):
        """
        Read object from DB given the key.
        If key has been read before, this call would simply
        return the cached value. Use reload to force reloading
        from DB.

        :param key: str: key to DB
        :param reload:  Optionally force reloading of the object
                        from db (Default value = False)
        :returns: The object from DB

example::

    obj = db.new('MyClass', key)
    obj.write()
    db.read(key) # read from cache
    db.read(key, reload=True) # Force loading from DB

        """
        path = self._get_full_path(key)
        res = None if reload else self._cache.get(path)
        if not res:
            res = self._deserialise(self.get_raw(path))
            if self.is_data_dbobj(res):
                res = self.read_dbobj(res)

        self._cache[key] = res
        return res

    def __setitem__(self, key: str, value):
        """
        Set data from the DB based on key

        :param key: str:  The key to set.
        :param value: str:  The python object

example::

    db[key] = value # sets key to value
        """
        self._cache[key] = value
        if self.is_dbobj(value):
            self.write_dbobj(value)
        else:
            self.set_raw(self._get_full_path(key), self._serialise(value))

    def get_raw(self, key: str):
        """
        Get data from the DB based on key.

        This is to be implemented by derived class.

        :param key: str:  The key to get, including base_path.
        :returns: str: The raw, pickled data.
        """
        raise NotImplementedError()

    def set_raw(self, key: str, value):
        """
        Set data from the DB based on key.

        This is to be implemented by derived class.

        :param key: str:  The key to set, including base_path.
        :param value: The raw, pickled data.
        """
        raise NotImplementedError()

    def delete_raw(self, key: str):
        """
        Delete data from the DB based on key

        This is to be implemented by derived class.

        :param key: str:  The key to get, including base_path.

        """
        raise NotImplementedError()

    def delete(self, key: str):
        """
        Delete data from the DB based on key

        :param key: str:  The key to get

example::

    db.delete(key) # Deletes data with key
        """
        del self._cache[key]
        self.delete_raw(self._get_full_path(key))

    def new(self, class_name: str, key: str, **kwargs):
        """
        Create a new object on the DB.
        The object is not persisted until obj.write() is called.

        :param class_name: str: name of the class.
                                This name must be in the config registry
        :param key: str: The key to persist on the DB
        :param kwargs: the stored attributes to set on the obj
        :returns: an obj of type defined by class_name

example::

    obj = db.new('MyClass', key, foo=3)
        """
        return self.db_obj_new(class_name, key, kwargs)

    def __repr__(self):
        """
        The representation of the db.

        i.e. <kydb.RedisDB redis://my-redis-host/source>
        """
        return f'<{type(self).__name__} {self.url}>'



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='kydb',
    packages=['kydb'],
    version='0.3.1',
    license='MIT',
    description='kydb (Kinyu Database). NoSQL DB interface.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Tony Yum',
    author_email='tony.yum@tayglobal.com',
    url='https://github.com/tayglobal/kydb',
    download_url='https://github.com/tayglobal/kydb/archive/v_01.tar.gz',
    keywords=['NoSQL', 'Database', 'DB'],
    install_requires=[            # I get to this in a second
        'boto3',
        'redis'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.6',
)
